from motorway.intersection import Intersection
import boto3
from utils import sqs_encode_to_json


class SQSInsertIntersection(Intersection):
    queue_name = None

    def __init__(self, **kwargs):
        super(SQSInsertIntersection, self).__init__(**kwargs)
        self.sqs = boto3.resource(**self.connection_parameters())
        assert self.queue_name, "Please define attribute queue_name on your SQSRamp"
        self.queue = self.sqs.create_queue(QueueName=self.queue_name)

    def connection_parameters(self):
        return {
            'region_name': 'eu-west-1',
            'service_name': 'sqs'
            # Add this or use ENV VARS
            # 'aws_access_key_id': '',
            # 'aws_secret_access_key': ''
        }

    def process(self, message):
        self.queue.send_message(MessageBody=sqs_encode_to_json(message.content))
        self.ack(message)
        yield



import decimal
from json import JSONEncoder
import datetime
from isodate import parse_duration, duration_isoformat, datetime_isoformat
import zmq
import socket

ramp_result_stream_name = lambda ramp_class_name: "_ramp_result_%s" % ramp_class_name


def percentile_from_dict(D, P):
    """
    Find the percentile of a list of values

    @parameter N - A dictionary, {observation: frequency}
    @parameter P - An integer between 0 and 100

    @return - The percentile of the values.
    """
    assert 0 < P <= 100, "Percentile must be in range (0, 100)"
    N = sum(D.values())
    P = float(P)/100
    n = int(N) * P

    i = 1  # The formula returns a 1-indexed number
    observation = None
    for observation in sorted(D.keys()):
        if i >= n:
            return observation
        i += D[observation]
    return observation
    # raise AssertionError("Didn't find percentile")


class DateTimeAwareJsonEncoder(JSONEncoder):
    def default(self, o):
        if type(o) == datetime.timedelta:
            return duration_isoformat(o)
        elif type(o) == datetime.datetime:
            return datetime_isoformat(o)
        elif isinstance(o, decimal.Decimal):
            return float(o)
        return super(DateTimeAwareJsonEncoder, self).default(o)


def set_timeouts_on_socket(scket):
    scket.RCVTIMEO = 10000
    scket.SNDTIMEO = 10000
    scket.LINGER = 1000


def get_connections_block(queue, refresh_connection_socket, limit=100, existing_connections=None):
    i = 0
    connections = existing_connections if existing_connections else {}
    while queue not in connections and i < limit:
        try:
            connections = refresh_connection_socket.recv_json()
        except zmq.Again:
            pass
        i += 1
    return connections


def get_ip():
    # From http://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 0))
        ip = s.getsockname()[0]
    except:
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip


import os
import time

from motorway.messages import Message
from motorway.ramp import Ramp

"""
Requires pip install recurly
"""


class RecurlyRamp(Ramp):
    @property
    def recurly(self):
        import recurly
        recurly.SUBDOMAIN = os.environ.get('RECURLY_SUBDOMAIN')
        recurly.API_KEY = os.environ.get('RECURLY_API_KEY')
        return recurly


class RecurlyInvoiceRamp(RecurlyRamp):

    def next(self):
        for invoice in self.recurly.Invoice.all():
            yield Message(invoice.uuid, {
                'uuid': invoice.uuid,
                'invoice_number': invoice.invoice_number,
                'vat_number': invoice.vat_number,
                'total_in_cents': invoice.total_in_cents,
                'tax_in_cents': invoice.tax_in_cents,
                'subtotal_in_cents': invoice.subtotal_in_cents,
                'state': invoice.state,
                'collection_method': invoice.collection_method,
                'currency': invoice.currency,
                'account': invoice.account().account_code,  # TODO: Parse this from the href in the XML
                'created_at': invoice.created_at,
                'updated_at': invoice.updated_at,
                'closed_at': invoice.closed_at,
            })
        time.sleep(3600)


class RecurlyAccountRamp(RecurlyRamp):

    def next(self):
        for account in self.recurly.Account.all():
            yield Message(account.account_code, {
                'company_name': account.company_name,
                'account_code': account.account_code,
            })
        time.sleep(3600)



import json
import time
import os
import requests

from motorway.messages import Message
from motorway.ramp import Ramp
from simple_salesforce import Salesforce
from simple_salesforce.api import SFType



class SalesforceStreamingObjectRamp(Ramp):
    topic = 'YourTopic'


    def __init__(self, *args, **kwargs):
        super(SalesforceStreamingObjectRamp, self).__init__(*args, **kwargs)
        self.sf = Salesforce(username=self.sf_username, password=self.sf_password, security_token=self.sf_token)
        self.session = requests.Session()  # Important, it needs cookies to be persistent

        connect = self.session.post("https://%s/cometd/37.0" % self.sf.sf_instance, headers=self.sf.headers,
                                    data=json.dumps({
                                        "version": "1.0",
                                        "minimumVersion": "0.9",
                                        "channel": "/meta/handshake",
                                        "supportedConnectionTypes": ["long-polling"],
                                        "advice": {
                                             "timeout": 2500,
                                             "interval": 0
                                        },
                                        "id": "1"
                                    }))

        self.connection = connect.json()[0]

        subscribe = self.session.post("https://%s/cometd/37.0" % self.sf.sf_instance, headers=self.sf.headers,
                                      data=json.dumps({
                                          "channel": "/meta/subscribe",
                                          "subscription": "/topic/%s" % self.topic,
                                          "clientId": self.connection['clientId']
                                      }))
        assert subscribe.json()[0]['successful']

    @property
    def sf_username(self):
        return os.environ.get("SALESFORCE_USERNAME")

    @property
    def sf_password(self):
        return os.environ.get("SALESFORCE_PASSWORD")

    @property
    def sf_token(self):
        return os.environ.get("SALESFORCE_TOKEN")

    def next(self):
        updates = self.session.post("https://%s/cometd/37.0" % self.sf.sf_instance, headers=self.sf.headers,
                           data=json.dumps({
                               "channel": "/meta/connect",
                               "connectionType": "long-polling",
                               "clientId": self.connection['clientId']
                           }))
        for update in updates.json():
            if 'data' in update:
                if 'sobject' in update['data']:
                    yield Message(update['data']['sobject']['Id'],
                        update['data']['sobject']
                    )


class SalesforceQueryObjectRamp(Ramp):
    query = 'SELECT Id, Amount FROM Opportunity LIMIT 10'
    sleep_time = 60 * 30  # Every 30 min


    def __init__(self, *args, **kwargs):
        super(SalesforceQueryObjectRamp, self).__init__(*args, **kwargs)
        self.sf = Salesforce(username=self.sf_username, password=self.sf_password, security_token=self.sf_token)

    @property
    def sf_username(self):
        return os.environ.get("SALESFORCE_USERNAME")

    @property
    def sf_password(self):
        return os.environ.get("SALESFORCE_PASSWORD")

    @property
    def sf_token(self):
        return os.environ.get("SALESFORCE_TOKEN")

    def next(self):
        for obj in self.sf.query_all(self.query)['records']:
            attributes = obj.pop('attributes')
            yield Message(
                obj['Id'],
                obj
            )
        time.sleep(self.sleep_time)



import calendar
import logging
import datetime
import socket

from motorway.intersection import Intersection
from motorway.messages import Message
from motorway.utils import set_timeouts_on_socket, get_ip
import time
import zmq

logger = logging.getLogger(__name__)

current_heartbeat = lambda: calendar.timegm(datetime.datetime.utcnow().timetuple())


class ConnectionIntersection(Intersection):
    """
    Responsible to receiving and publishing information about the different message endpoints. Every ramp/intersection
    will subscribe to this information and use it for routing the messages correctly.

    While implemented with the intersection interface, it is currently only possibly to run this as a single instance
    as it has internal state which is not shared.

    It should be possible to extend this class and override it to store state in external systems such as Consul,
    Zookeper or something else.
    """

    HEARTBEAT_TIMEOUT = 30

    send_control_messages = False

    def __init__(self, bind_address='tcp://0.0.0.0:7007', **kwargs):
        super(ConnectionIntersection, self).__init__(**kwargs)
        self.process_statistics = {}
        self.ramp_socks = {}
        self.process_address_to_uuid = {}
        self.queue_processes = {}
        self.process_id_to_name = {}
        self.bind_address = bind_address

    def process(self, message):
        connection_updates = message.content
        self.process_id_to_name[connection_updates['meta']['id']] = connection_updates['meta']['name']
        for queue, consumers in connection_updates['streams'].items():
            if queue not in self.queue_processes:
                self.queue_processes[queue] = {
                    'streams': [],
                    'stream_heartbeats': {},
                    'grouping': None
                }
            for consumer in consumers:
                if consumer not in self.queue_processes[queue]['streams']:
                    # self.process_statistics[connection_updates['meta']['id']]['status'] = 'running'
                    self.process_address_to_uuid[consumer] = connection_updates['meta']['id']
                    # Add to streams and update grouping TODO: Keep track of multiple groupings
                    self.queue_processes[queue]['streams'].append(consumer)
                    self.queue_processes[queue]['grouping'] = connection_updates['meta']['grouping']
                self.queue_processes[queue]['stream_heartbeats'][consumer] = {
                    'heartbeat': current_heartbeat(),
                    'process_name': connection_updates['meta']['name'],
                    'process_id': connection_updates['meta']['id']
                }
                logger.debug("Received heartbeat from %s on queue %s - current value %s" % (consumer, queue, self.queue_processes[queue]['stream_heartbeats'][consumer]))
        yield

    def connection_thread(self, context=None, **kwargs):
        while not self.receive_port:
            time.sleep(1)

        # The publish/broadcast socket where clients subscribe to updates
        broadcast_connection_sock = context.socket(zmq.PUB)
        broadcast_connection_sock.bind(self.bind_address)
        set_timeouts_on_socket(broadcast_connection_sock)

        self.queue_processes['_update_connections'] = {
            'streams': ['tcp://%s:%s' % (get_ip(), self.receive_port)],
            'grouping': None,
            'stream_heartbeats': {}
        }
        while True:
            for queue, consumers in self.queue_processes.items():
                for consumer, heartbeat_info in consumers['stream_heartbeats'].items():
                    if current_heartbeat() > (heartbeat_info['heartbeat'] + self.HEARTBEAT_TIMEOUT):
                        logger.warn("Removing %s from %s due to missing heartbeat" % (consumer, queue))
                        self.queue_processes[queue]['streams'].remove(consumer)
                        self.queue_processes[queue]['stream_heartbeats'].pop(consumer)
                        # self.process_statistics[heartbeat_info['process_id']]['status'] = 'failed'

            # Send the current connections
            broadcast_connection_sock.send_json(self.queue_processes)
            logger.debug("Announced %s", self.queue_processes)

            # This intersection should it self publish to send socks..
            # self.set_send_socks(self.queue_processes, self.output_queue, self.context)
            time.sleep(5)


"""
You need https://pypi.python.org/pypi/simple-salesforce
"""
import os

import requests
from simple_salesforce import SFType
from simple_salesforce import Salesforce

from motorway.intersection import Intersection


class SalesforceInsertIntersection(Intersection):
    upsert = True
    object_api_name = None
    upsert_id_field = None

    SF_USERNAME = os.environ.get('SF_USERNAME')
    SF_PASSWORD = os.environ.get('SF_PASSWORD')
    SF_SECURITY_TOKEN = os.environ.get('SF_SECURITY_TOKEN')

    def __init__(self, *args, **kwargs):
        super(SalesforceInsertIntersection, self).__init__(*args, **kwargs)
        session = requests.Session()
        self.sf_instance = Salesforce(
            username=self.SF_USERNAME,
            password=self.SF_PASSWORD,
            security_token=self.SF_SECURITY_TOKEN,
            session=session
        )
        self.sf_type = SFType(self.object_api_name, self.sf_instance.session_id, self.sf_instance.sf_instance, self.sf_instance.sf_version, self.sf_instance.proxies)

    def process(self, message):
        if self.upsert:
            self.sf_type.upsert(
                '%s/%s' % (self.upsert_id_field, message.content.pop(self.upsert_id_field)),
                message.content
            )
        else:
            self.sf_type.create(message.content)
        self.ack(message)


from sqlalchemy import MetaData, Table, create_engine, Integer, Column, select, bindparam
from sqlalchemy.exc import NoSuchTableError
from motorway.decorators import batch_process
from motorway.intersection import Intersection


class DatabaseInsertIntersection(Intersection):
    """
    Excepts messages with a single dictionary where the key is the field name in the database and the value is .. value
    """
    database_uri = None
    schema = 'public'
    table = None
    create_table = True
    add_columns = False
    batch_size = 100
    upsert = True
    table_columns = [
        {'name': 'id', 'type': Integer, 'primary_key': True}
    ]

    debug_sql_statements = False

    def __init__(self, *args, **kwargs):
        self.engine = create_engine(self.database_uri, echo=self.debug_sql_statements)
        self.metadata = MetaData(bind=self.engine)
        self.connection = self.engine.connect()
        try:
            self.table = Table(self.table, self.metadata, autoload=True, schema=self.schema)
        except NoSuchTableError:
            if not self.create_table:
                raise
            else:
                self.table = Table(self.table, self.metadata,
                    *[
                        Column(column['name'], column['type'], primary_key=column.get('primary_key', False))
                    for column in self.table_columns],
                schema = self.schema)
                self.table.create(self.engine)

        super(DatabaseInsertIntersection, self).__init__(*args, **kwargs)

    @staticmethod
    def _column_name(name):
        return 'b_%s' % name

    @staticmethod
    def bind_alias(dct):
        return {DatabaseInsertIntersection._column_name(key): value for key, value in dct.items()}

    @batch_process(limit=batch_size)  # todo: make this work with inheritance
    def process(self, messages):
        if not self.upsert:  # Just insert
            self.connection.execute(self.table.insert(), [
                message.content for message in messages
            ])
        else:  # "Upsert"
            primary_key = [col['name'] for col in self.table_columns if col.get('primary_key')]
            assert len(primary_key) == 1, "More than one primary key defined"
            primary_key = primary_key[0]  # Assume there is always only one
            pk_field = getattr(self.table.c, primary_key)
            messages_by_pk = {message.content.get(primary_key): message for message in messages}


            ##### GET PKS FOR UPDATE #####
            update_pks = [row[0] for row in self.connection.execute(
                select(
                    [pk_field],
                    pk_field.in_(
                        messages_by_pk.keys()
                    )
                )
            ).fetchall()]

            #### ACTUAL UPDATE OF NON-PK FIELDS IN ROWS ####
            if len(self.table_columns) > 1:  # No need to update if PK is the only column, in which case its already correct

                stmt = self.table.update().where(
                    pk_field == bindparam(self._column_name(primary_key))
                ).values(
                    {
                        col['name']: bindparam(self._column_name(col['name'])) for col in self.table_columns if not col.get('primary_key')
                    }
                )


                if update_pks:
                    self.connection.execute(stmt, [
                        self.bind_alias(message.content) for message_pk, message in messages_by_pk.items() if message_pk in update_pks
                    ])

            #### BULK INSERT ####

            # Compare number of returned existing rows with unique rows
            if not len(update_pks) == len(messages_by_pk.keys()):
                self.connection.execute(self.table.insert(), [
                    message.content for message_pk, message in messages_by_pk.items() if message_pk not in update_pks
                ])

        for message in messages:
            self.ack(message)

        yield


import datetime
import json
import logging
from threading import Thread
from flask import Flask, render_template, Response
from isodate import parse_duration
from motorway.intersection import Intersection
from motorway.utils import DateTimeAwareJsonEncoder

logger = logging.getLogger(__name__)


class WebserverIntersection(Intersection):
    """
    Simple Flask webserver that exposes information sent from the controller(s).

    It groups processes by name and furhter more sets a "status" for the different processes, based on
    how busy they are.
    """

    send_control_messages = False

    def get_last_minutes(self, minute_count=10):
        now = datetime.datetime.now()
        return [(now - datetime.timedelta(minutes=i)).minute for i in range(0, minute_count)]

    def __init__(self, *args, **kwargs):
        super(WebserverIntersection, self).__init__(*args, **kwargs)

        self.process_statistics = {}
        self.stream_consumers = {}
        self.failed_messages = {}
        self.groups = {}

        app = Flask(__name__)

        @app.route("/")
        def index():
            return render_template("index.html")

        @app.route("/app.js")
        def js():
            return Response(render_template("app.js"), mimetype='application/javascript')

        @app.route("/style.css")
        def css():
            return Response(render_template("style.css"), mimetype='text/css')

        @app.route("/detail/<process>/")
        def detail(process):
            return render_template(
                "detail.html",
                process=process,
                failed_messages=reversed(sorted([msg for msg in self.failed_messages.values() if
                                 msg[1] == process], key=lambda itm: itm[0])[-20:]),

            )

        @app.route("/api/status/")
        def api_status():
            return Response(json.dumps(dict(
                groups=self.groups,
                last_minutes=self.get_last_minutes()
            ), cls=DateTimeAwareJsonEncoder), mimetype='application/json', headers={
                'Access-Control-Allow-Origin': '*'
            })

        # @app.route("/api/detail/<process>/")
        # def api_detail(process):
        #     return Response(json.dumps(dict(
        #         failed_messages=[msg[1] for msg in self.failed_messages.values() if self.process_id_to_name[msg[0]] == process],
        #     ), cls=DateTimeAwareJsonEncoder), mimetype='application/json')

        p = Thread(target=app.run, name="motorway-webserver", kwargs=dict(
            port=5000,
            host="0.0.0.0",
        ))
        p.start()

    def process(self, message):
        self.process_statistics = message.content['process_statistics']
        self.stream_consumers = message.content['stream_consumers']
        self.failed_messages = message.content['failed_messages']

        for process_id, stats in self.process_statistics.items():
            if process_id in self.process_id_to_name:

                stats['state'] = 'available'
                if stats['waiting'] > 0:
                    stats['state'] = 'busy'
                    processed_last_few_minutes = 0
                    for minute in self.get_last_minutes(3):
                        processed_last_few_minutes += stats['histogram'][str(minute)]['processed_count']
                    if stats['waiting'] > processed_last_few_minutes:
                        stats['state'] = 'overloaded'

                group_name = self.process_id_to_name[process_id].split('-')[0]
                new_stats = self.groups.get(group_name, {'processes': {}.copy()}.copy())
                new_stats['processes'][process_id] = stats
                self.groups[group_name] = new_stats

        for group in self.groups.values():
            group['waiting'] = sum([process['waiting'] for process in group['processes'].values()])
            group['time_taken'] = datetime.timedelta()
            group['histogram'] = {str(minute): {'error_count': 0, 'success_count': 0, 'timeout_count': 0, 'processed_count': 0}.copy() for minute in range(0, 60)}.copy()
            for process_id, process in group['processes'].items():

                # Remove stale processes (those no longer in the connection thread)
                if process_id not in self.process_id_to_name:
                    del group['processes'][process_id]
                    continue

                # Calculate statistics on the active processes
                group['time_taken'] += parse_duration(process['time_taken']) or datetime.timedelta(seconds=0)
                for minute, histogram_dict in process.get('histogram').items():
                    group['histogram'][minute]['error_count'] += histogram_dict['error_count']
                    group['histogram'][minute]['success_count'] += histogram_dict['success_count']
                    group['histogram'][minute]['timeout_count'] += histogram_dict['timeout_count']
                    group['histogram'][minute]['processed_count'] += histogram_dict['processed_count']
            group['frequency'] = sum([sum(process['frequency'].values()) for process in group['processes'].values()]) or 1  # Fallback to at least one, otherwise division fails below

            group['avg_time_taken'] = group['time_taken'] / group['frequency'] / len(group['processes']) if len(group['processes']) else 0

        yield


import random
from Queue import Queue
import json
from threading import Thread, Lock, Semaphore
import time
import uuid
import datetime
import logging
import boto3
from motorway.messages import Message
from motorway.ramp import Ramp
from boto3.dynamodb.conditions import Attr
shard_election_logger = logging.getLogger("motorway.contrib.amazon_kinesis.shard_election")

logger = logging.getLogger(__name__)


class NoItemsReturned(Exception):
    pass


class KinesisRamp(Ramp):
    stream_name = None
    heartbeat_timeout = 30  # Wait 10 seconds for a heartbeat update, or kill it
    MAX_UNCOMPLETED_ITEMS = 3000
    GET_RECORDS_LIMIT = 1000

    def __init__(self, shard_threads_enabled=True, **kwargs):
        super(KinesisRamp, self).__init__(**kwargs)
        self.conn = boto3.client(**self.connection_parameters('kinesis'))
        assert self.stream_name, "Please define attribute stream_name on your KinesisRamp"

        control_table_name = self.get_control_table_name()

        self.worker_id = str(uuid.uuid4())
        self.semaphore = Semaphore()
        self.uncompleted_ids = {}
        self.dynamodb_client = boto3.client(**self.connection_parameters('dynamodb'))
        if shard_threads_enabled:
            self.dynamodb = boto3.resource(**self.connection_parameters('dynamodb'))


            try:
                self.dynamodb_client.describe_table(TableName=control_table_name)
            except self.dynamodb_client.exceptions.ResourceNotFoundException:
                self.dynamodb_client.create_table(
                    TableName=control_table_name,
                    KeySchema=[
                        {
                            'AttributeName': 'shard_id',
                            'KeyType': 'HASH'
                        },
                    ],
                    ProvisionedThroughput={
                        'ReadCapacityUnits': 10,
                        'WriteCapacityUnits': 10
                    },
                    AttributeDefinitions=[
                        {
                            'AttributeName': 'shard_id',
                            'AttributeType': 'S'
                        },
                    ],
                )
                self.dynamodb_client.get_waiter('table_exists').wait(TableName=control_table_name)

            self.control_table = self.dynamodb.Table(control_table_name)
            shards = self.conn.describe_stream(StreamName=self.stream_name)['StreamDescription']['Shards']
            random.shuffle(shards)  # Start the threads in random order, in case of bulk restart
            threads = []
            self.insertion_queue = Queue()
            for i, shard in enumerate(shards):
                self.uncompleted_ids[shard['ShardId']] = set()
                t = Thread(target=self.process_shard, name="%s-%s" % (self.__class__.__name__, i), args=(shard['ShardId'], ))
                threads.append(t)
                t.start()

    def get_control_table_name(self):
        return 'pipeline-control-%s' % self.stream_name

    def claim_shard(self, shard_id):
        shard_election_logger.info("Claiming shard %s" % shard_id)
        try:
            control_record = self.control_table.get_item(Key={'shard_id': shard_id})['Item']
        except KeyError:
            raise NoItemsReturned()
        control_record['worker_id'] = self.worker_id
        control_record['heartbeat'] = 0
        try:
            self.control_table.put_item(Item=control_record,
                                        ConditionExpression=Attr('shard_id').eq(shard_id) & Attr('checkpoint').eq(control_record['checkpoint'])
                                        # ensure that the record was not changed between the get and put.
                                        )
        except self.dynamodb_client.exceptions.ConditionalCheckFailedException:  # Someone else edited the record
            shard_election_logger.debug("Failed to claim %s to %s" % (shard_id, self.worker_id))
            return False
        return True

    def can_claim_shard(self, shard_id):
        try:
            control_record = self.control_table.get_item(Key={'shard_id': shard_id})['Item']
        except KeyError:
            raise NoItemsReturned()
        heartbeat = control_record['heartbeat']
        time.sleep(self.heartbeat_timeout)
        if heartbeat == self.control_table.get_item(Key={'shard_id': shard_id})['Item']['heartbeat']:  # TODO: check worker_id as well
            shard_election_logger.debug("Shard %s - heartbeat remained unchanged for defined time, taking over" % shard_id)
            return True
        else:
            shard_election_logger.debug("Shard %s - Heartbeat changed, continue sleeping" % shard_id)
        # Balance, if possible
        active_workers = {
            self.worker_id: True
        }
        heartbeats = {
            self.worker_id: 0
        }

        shards = list(self.control_table.scan())
        for shard in shards:  # Update active worker cache
            if shard['worker_id'] not in active_workers:
                heartbeats[shard['worker_id']] = shard['heartbeat']
            if heartbeats[shard['worker_id']] == shard['heartbeat']:
                active_workers[shard['worker_id']] = False
            else:
                active_workers[shard['worker_id']] = True
        number_of_active_workers = sum([1 for is_active in active_workers.values() if is_active])
        number_of_shards = len(shards)
        optimal_number_of_shards_per_worker = number_of_shards / number_of_active_workers
        workers = set([shard['worker_id'] for shard in shards])
        shards_per_worker = {worker: sum([1 for shard in shards if shard['worker_id'] == worker]) for worker in workers}
        for shard in shards:
            if shard['shard_id'] == shard_id:
                if (
                    # Check if the shards current worker has too many, or if the worker has no workers, then take
                    # the shard if the current worker has more than one shard!
                        shards_per_worker.get(shard['worker_id'], 0) > optimal_number_of_shards_per_worker or (
                                    shards_per_worker.get(self.worker_id, 0) == 0 and
                                    shards_per_worker.get(shard['worker_id'], 0) > 1
                                )
                ) and (
                    # Only get shards for balancing purposes, if we have too little
                        shards_per_worker.get(self.worker_id, 0) < optimal_number_of_shards_per_worker
                ):
                        shard_election_logger.debug("Taking over %s from %s" % (shard_id, shard['worker_id']))
                        return True
        return False

    def process_shard(self, shard_id):
        while True:
            try:
                # try to claim the shard
                try:
                    while True:
                        with self.semaphore:
                            if self.can_claim_shard(shard_id):
                                if self.claim_shard(shard_id):
                                    break
                        time.sleep(random.randrange(2, 15))
                except NoItemsReturned:
                    # no record for this shard found, nobody ever claimed the shard yet, so claim it
                    self.control_table.put_item(Item={
                        'shard_id': shard_id,
                        'checkpoint': 0,
                        'worker_id': self.worker_id,
                        'heartbeat': 0,
                    })

                # get initial iterator
                control_record = self.control_table.get_item(Key={'shard_id': shard_id})['Item']
                if control_record['checkpoint'] > 0:
                    # if we have a checkpoint, start from the checkpoint
                    iterator = self.conn.get_shard_iterator(
                        StreamName=self.stream_name,
                        ShardId=shard_id,
                        ShardIteratorType="AT_SEQUENCE_NUMBER",
                        StartingSequenceNumber=str(control_record['checkpoint'])
                    )['ShardIterator']
                else:
                    # we have no checkpoint stored, start from the latest item in Kinesis
                    iterator = self.conn.get_shard_iterator(
                        StreamName=self.stream_name,
                        ShardId=shard_id,
                        ShardIteratorType="LATEST",
                    )['ShardIterator']

                cloudwatch = boto3.client(**self.connection_parameters('cloudwatch'))
                current_minute = lambda: datetime.datetime.now().minute
                minute = None
                latest_item = None
                while True:
                    control_record = self.control_table.get_item(Key={'shard_id': shard_id})['Item']  # always retrieve this at the top of the loop
                    current_checkpoint = control_record['checkpoint']
                    current_heartbeat = control_record['heartbeat']
                    # if the shard was claimed by another worker, break out of the loop
                    if not control_record['worker_id'] == self.worker_id:
                        shard_election_logger.info("Lost shard %s, going back to standby" % shard_id)
                        break

                    # update the heartbeat and the checkpoint
                    control_record['heartbeat'] += 1
                    if len(self.uncompleted_ids[shard_id]):
                        # Get the "youngest" uncompleted sequence number
                        control_record['checkpoint'] = min(self.uncompleted_ids[shard_id])
                    elif latest_item:
                        # or the latest item we yielded
                        control_record['checkpoint'] = latest_item

                    self.control_table.put_item(Item=control_record,
                                                ConditionExpression=Attr('shard_id').eq(shard_id) & Attr('checkpoint').eq(current_checkpoint) & Attr('worker_id').eq(self.worker_id) & Attr('heartbeat').eq(current_heartbeat)
                                                # Will fail if someone else modified it - ConditionalCheckFailedException
                                                )

                    if len(self.uncompleted_ids[shard_id]) < self.MAX_UNCOMPLETED_ITEMS:
                        # get records from Kinesis, using the previously created iterator
                        result = self.conn.get_records(ShardIterator=iterator, Limit=self.GET_RECORDS_LIMIT)

                        # insert the records into the queue, and use the provided iterator for the next loop
                        for record in result['Records']:
                            self.uncompleted_ids[shard_id].add(record['SequenceNumber'])
                            latest_item = record['SequenceNumber']
                            self.insertion_queue.put(record)
                        iterator = result['NextShardIterator']
                    else:
                        logger.debug("Pausing, too many uncompleted items (%s/%s)" % (len(self.uncompleted_ids[shard_id]), self.MAX_UNCOMPLETED_ITEMS))
                        # we have too many uncompleted items, so back off for a while
                        # however, the iterator needs to be updated, because it expires after a while
                        # use the latest record we added to the queue as the starting point
                        next_iterator_number = latest_item if latest_item else str(control_record['checkpoint'])
                        iterator = self.conn.get_shard_iterator(
                            StreamName=self.stream_name,
                            ShardId=shard_id,
                            ShardIteratorType="AT_SEQUENCE_NUMBER",
                            StartingSequenceNumber=next_iterator_number
                        )['ShardIterator']

                        # get just one item to update the MillisBehindLatest below
                        result = self.conn.get_records(ShardIterator=iterator, Limit=1)

                    # Push metrics to CloudWatch
                    delay = result['MillisBehindLatest']
                    if minute != current_minute():  # push once per minute to CloudWatch.
                        minute = current_minute()
                        cloudwatch.put_metric_data(Namespace='Motorway/Kinesis',
                                                   MetricData=[{'MetricName': 'MillisecondsBehind',
                                                                'Dimensions': [{
                                                                    'Name': 'Stream',
                                                                    'Value': self.stream_name
                                                                }, {
                                                                    'Name': 'Shard',
                                                                    'Value': shard_id
                                                                }],
                                                                'Value': delay,
                                                                'Unit': 'Milliseconds'
                                                                }])

                    # recommended pause between fetches from AWS
                    time.sleep(1)
            except self.dynamodb_client.exceptions.ConditionalCheckFailedException as e:
                logger.warning(e)
                pass  # we're no longer worker for this shard
            except (self.dynamodb_client.exceptions.ProvisionedThroughputExceededException, self.dynamodb_client.exceptions.ProvisionedThroughputExceededException,
                    self.conn.exceptions.LimitExceededException, self.conn.exceptions.ProvisionedThroughputExceededException):
                logger.warning(e)
                time.sleep(random.randrange(5, self.heartbeat_timeout/2))  # back off for a while

    def connection_parameters(self, service_name):
        return {
            'region_name': 'eu-west-1',
            'service_name': service_name,
            # Add this or use ENV VARS
            # 'aws_access_key_id': '',
            # 'aws_secret_access_key': ''
        }

    def next(self):
        msg = self.insertion_queue.get()
        try:
            yield Message(msg['SequenceNumber'], json.loads(msg['Data']), grouping_value=msg['PartitionKey'])
        except ValueError as e:
            logger.exception(e)

    def success(self, _id):
        for uncompleted_ids in self.uncompleted_ids.values():
            if _id in uncompleted_ids:
                uncompleted_ids.remove(_id)



from collections import defaultdict
import random
import time
import uuid
from motorway.contrib.amazon_sqs.intersections import SQSInsertIntersection
from motorway.decorators import batch_process
from motorway.messages import Message
from motorway.intersection import Intersection


class SentenceSplitIntersection(Intersection):
    def process(self, message):
        for word in message.content.split(" "):
            yield Message(str(uuid.uuid4()), word, grouping_value=word)
        self.ack(message)


class WordCountIntersection(Intersection):
    def __init__(self, **kwargs):
        self._count = defaultdict(int)
        super(WordCountIntersection, self).__init__(**kwargs)

    @batch_process(wait=2, limit=5)
    def process(self, messages):
        for message in messages:
            # time.sleep(1)
            self._count[message.content] += 1
            self.ack(message)
        # time.sleep(0.5)
        yield Message(str(uuid.uuid4()), self._count)


class AggregateIntersection(Intersection):
    def __init__(self, **kwargs):
        self._count = defaultdict(int)
        super(AggregateIntersection, self).__init__(**kwargs)

    def process(self, message):
        self._count.update(message.content)
        yield Message(str(uuid.uuid4()), self._count)
        self.ack(message)


class AggregateConsumerIntersection(Intersection):
    def __init__(self, **kwargs):
        self._count = defaultdict(int)
        super(AggregateConsumerIntersection, self).__init__(**kwargs)

    def process(self, message):
        self._count.update(message.content)
        self.ack(message)
        print "Aggregate: ", self.process_uuid, self._count
        yield



class ExampleSQSIntersection(SQSInsertIntersection):
    queue_name = "tutorial_motorway"


import time
import uuid
from motorway.contrib.amazon_kinesis.ramps import KinesisRamp
from motorway.contrib.amazon_kinesis.intersections import KinesisInsertIntersection
from motorway.contrib.amazon_sqs.ramps import SQSRamp
from motorway.messages import Message
from motorway.ramp import Ramp
import random


class WordRamp(Ramp):
    sentences = [
        "Oak is strong and also gives shade.",
        "Cats and dogs each hate the other.",
        "The pipe began to rust while new.",
        "Open the crate but don't break the glass.",
        "Add the sum to the product of these three.",
        "Thieves who rob friends deserve jail.",
        "The ripe taste of cheese improves with age.",
        "Act on these orders with great speed.",
        "The hog crawled under the high fence.",
        "Move the vat over the hot fire.",
    ]

    def __init__(self, *args, **kwargs):
        super(WordRamp, self).__init__(*args, **kwargs)
        self.limit = 10000
        self.progress = 1

    def next(self):
        # yield Message(uuid.uuid4().int, self.sentences[random.randint(0, len(self.sentences) -1)])
        if self.progress <= self.limit:
            self.progress += 1
            # time.sleep(10)
            sentence = self.sentences[random.randint(0, len(self.sentences) -1)]
            yield Message(uuid.uuid4().int, sentence, grouping_value=sentence)
        else:
            time.sleep(1)

    def success(self, _id):
        pass
        #print "WordRamp %s was successful" % _id

    def failed(self, _id):
        print "WordRamp %s has failed" % _id

    def should_run(self):
        return True


class ExampleSQSRamp(SQSRamp):
    queue_name = "tutorial_motorway"


class ExampleKinesisRamp(KinesisRamp):
    stream_name = "data-pipeline-test"


class ExampleKinesisIntersection(KinesisInsertIntersection):
    stream_name = "data-pipeline-test"



import logging

from motorway.pipeline import Pipeline
from examples.ramps import WordRamp, ExampleSQSRamp
from examples.intersections import SentenceSplitIntersection, WordCountIntersection, ExampleSQSIntersection


logging.basicConfig(level=logging.INFO)


class WordCountPipeline(Pipeline):
    def definition(self):
        self.add_ramp(WordRamp, 'sentence')
        self.add_intersection(SentenceSplitIntersection, 'sentence', 'word_to_sqs')
        self.add_intersection(ExampleSQSIntersection, 'word_to_sqs')

        self.add_ramp(ExampleSQSRamp, 'word_from_sqs')
        self.add_intersection(WordCountIntersection, 'word_from_sqs')


WordCountPipeline().run()


import json
import boto3
from motorway.decorators import batch_process
from motorway.intersection import Intersection
from time import sleep
import logging
logger = logging.getLogger(__name__)


class KinesisInsertIntersection(Intersection):
    stream_name = None

    def __init__(self, **kwargs):
        super(KinesisInsertIntersection, self).__init__(**kwargs)
        self.conn = boto3.client(**self.connection_parameters())
        assert self.stream_name, "Please define attribute stream_name on your KinesisInsertIntersection"

    def connection_parameters(self):
        return {
            'region_name': 'eu-west-1',
            'service_name': 'kinesis',
            # Add this or use ENV VARS
        }

    @batch_process(limit=500, wait=1)
    def process(self, messages):
        """
        wait 1 second and get up to 500 items 
        Each PutRecords request can support up to 500 records. 
        Each record in the request can be as large as 1 MB, up to a limit of 5 MB for the entire request, including partition keys. 
        Each shard can support writes up to 1,000 records per second, up to a maximum data write total of 1 MB per second.
        This means we can run 2 intersections (2 x 500 records) submitting to the same shard before hitting the write limit (1000 records/sec)
        If we hit the write limit we wait 2 seconds and try to send the records that failed again, rinse and repeat
        If any other error than ProvisionedThroughputExceededException (only InternalFailure is currently supported) is returned in the response
        we log it using loglevel error and dump the message for replayability instead of raising an exception that would drop the whole batch.
        So if you are going to use this intersection in production be sure to monitor and handle the messages with log level error!
        :param messages: 
        :return: 
        """
        records = []
        for message in messages:
            kinesis_record = {'PartitionKey': message.grouping_value, 'Data': json.dumps(message.content)}
            records.append(kinesis_record)

        while len(records) > 0:
            logger.debug('put %s records' % len(records))
            response = self.conn.put_records(Records=records, StreamName=self.stream_name)
            records = []
            for i, record in enumerate(response['Records']):
                if len(record.get('ErrorCode', '')) > 0:
                    if record['ErrorCode'] == 'ProvisionedThroughputExceededException':
                        # retry when throttled
                        logger.warning(record['ErrorCode'])
                        kinesis_record = {'PartitionKey': messages[i].grouping_value, 'Data': json.dumps(messages[i].content)}
                        records.append(kinesis_record)
                    else:
                        # fail on any other exception
                        logger.error(record['ErrorMessage'])
                        self.fail(messages[i])
                else:
                    # success
                    self.ack(messages[i])

            if len(records) > 0:
                # wait to not get throttled again
                logger.warning('%s records were throttled, sleeping and re-trying...' % len(records))
                sleep(1)

        yield



import logging
import multiprocessing
from setproctitle import setproctitle
import time
import zmq

logger = logging.getLogger(__name__)


class ThreadRunner(object):

    def __init__(self):
        super(ThreadRunner, self).__init__()

    def thread_factory(self, *args, **kwargs):
        raise NotImplementedError("Define the threads your want to the ThreadRunner to manage in self.thread_factory()")

    @classmethod
    def run(cls, *args, **kwargs):
        self = cls(process_uuid=kwargs.get('process_uuid'))
        process_name = multiprocessing.current_process().name
        setproctitle("data-pipeline: %s" % process_name)

        factories = {factory: None for factory in self.thread_factory(*args, **kwargs)}

        logger.info("Running %s" % process_name)
        # Run all thread factories
        for factory in factories.keys():
            factories[factory] = factory()
            factories[factory].start()

        # Monitor and restart factories if needed
        while True:
            for factory, thread in factories.items():
                if not thread.isAlive():
                    logger.warn("Thread %s crashed in %s. Restarting" % (factory(), self.__class__.__name__))
                    factories[factory] = factory()
                    factories[factory].start()
            time.sleep(10)


from examples.intersections import SentenceSplitIntersection, AggregateIntersection, WordCountIntersection
from examples.ramps import WordRamp
from motorway.contrib.sql_alchemy.intersections import DatabaseInsertIntersection
from motorway.intersection import Intersection
from motorway.messages import Message
from motorway.pipeline import Pipeline
import sqlalchemy
import logging
import logging.config


logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'colored': {
                '()': 'colorlog.ColoredFormatter',
                'format': "%(log_color)s%(levelname)-8s%(reset)s %(name)-32s %(processName)-32s %(blue)s%(message)s"
        }
    },
    'loggers': {
        'motorway': {
            'level': 'WARN',
            'handlers': ['console'],
            'propagate': False
        },
        'werkzeug': {
            'level': 'WARN',
            'handlers': ['console'],
            'propagate': False,
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'colored'
        }
    },
    'root': {
        'level': 'WARN',
        'handlers': ['console']
    }

})



class SampleDatabaseInsertIntersection(DatabaseInsertIntersection):
    table = 'sample_tbl'
    database_uri = 'postgresql://localhost:5432/sampledb'

    table_columns = [
        # {'name': 'id', 'type': sqlalchemy.Integer, 'primary_key': True},
        {'name': 'word', 'type': sqlalchemy.String, 'primary_key': True},
        {'name': 'count', 'type': sqlalchemy.Integer},
    ]


class WordToDictIntersection(Intersection):
    def process(self, message):
        for word, count in message.content.items():
            yield Message.new(message, {
                'word': word,
                'count': count
            })
        self.ack(message)


class SamplePipeline(Pipeline):
    def definition(self):
        self.add_ramp(WordRamp, 'sentence')
        self.add_intersection(SentenceSplitIntersection, 'sentence', 'word')
        self.add_intersection(WordCountIntersection, 'word', 'formatter')
        self.add_intersection(WordToDictIntersection, 'formatter', 'insert_db')
        self.add_intersection(SampleDatabaseInsertIntersection, 'insert_db')


SamplePipeline().run()


import json
from motorway.contrib.amazon_sqs.exceptions import SQSMessageToLarge

SQS_MESSAGE_MAXIMUM_SIZE = 262144  # The maximum is 262,144 bytes (256 KB).
# https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/limits-messages.html


def sqs_encode_to_json(value, encoder_class=None):
    s = json.dumps(value, cls=encoder_class)
    message_size = len(s.encode('utf-8'))
    if message_size > SQS_MESSAGE_MAXIMUM_SIZE:
        raise SQSMessageToLarge('Message size of %sB exceeds the maximum SQS message size of %sB' % (message_size, SQS_MESSAGE_MAXIMUM_SIZE))
    return s



from motorway.pipeline import Pipeline


class ControllerPipeline(Pipeline):
    def definition(self):
        pass


ControllerPipeline(run_controller=True, run_webserver=False, run_connection_discovery=False, controller_bind_address="connections:7007").run()


from motorway.messages import Message
from motorway.ramp import Ramp
import boto3
import json


class SQSRamp(Ramp):
    queue_name = None
    json_group_key = None

    def __init__(self, *args, **kwargs):
        super(SQSRamp, self).__init__(*args, **kwargs)
        self.sqs = boto3.resource(**self.connection_parameters())
        assert self.queue_name, "Please define attribute queue_name on your SQSRamp"
        self.queue = self.sqs.create_queue(QueueName=self.queue_name)
        self.messages = {}

    def connection_parameters(self):
        return {
            'region_name': 'eu-west-1',
            'service_name': 'sqs'
            # Add this or use ENV VARS
            # 'aws_access_key_id': '',
            # 'aws_secret_access_key': ''
        }

    def next(self):
        for msg in self.queue.receive_messages(MaxNumberOfMessages=10, WaitTimeSeconds=5, VisibilityTimeout=10*60):
            # Gets max 10 messages, waiting for max 5 seconds to receive them and blocks other from receiving it for 10m
            self.messages[msg.message_id] = msg.receipt_handle  # we need the receipt_handle to delete the message and we always want to store the latest one, according to the sqs docs
            body = json.loads(msg.body)
            if self.json_group_key:
                yield Message(msg.message_id, body, grouping_value=body[self.json_group_key])
            else:
                yield Message(msg.message_id, body)

    def success(self, _id):
        if _id in self.messages:
            self.queue.delete_messages(Entries=[{'Id': _id, 'ReceiptHandle': self.messages[_id]}])  # TODO: Do this on ack
            del self.messages[_id]



# encoding: utf-8
from setuptools import setup, find_packages
from motorway import version

setup(
    name = 'motorway',
    version = version,
    description = 'Real-time ZMQ-based pure python data pipeline',
    author = u'Kristian Øllegaard',
    author_email = 'kristian@plecto.com',
    zip_safe=False,
    include_package_data = True,
    packages = find_packages(exclude=[]),
    install_requires=[
        open("requirements.txt").readlines(),
    ],
)


from motorway.grouping import HashRingGrouper

from motorway.pipeline import Pipeline
from examples.ramps import WordRamp, ExampleSQSRamp, ExampleKinesisIntersection, ExampleKinesisRamp
from examples.intersections import SentenceSplitIntersection, WordCountIntersection, AggregateIntersection
import logging.config

logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'colored': {
                '()': 'colorlog.ColoredFormatter',
                'format': "%(log_color)s%(levelname)-8s%(reset)s %(name)-32s %(processName)-32s %(blue)s%(message)s"
        }
    },
    'loggers': {
        'motorway': {
            'level': 'DEBUG',
            'handlers': ['console'],
            'propagate': False
        },
        'werkzeug': {
            'level': 'WARN',
            'handlers': ['console'],
            'propagate': False,
        },

    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'colored'
        }
    },
    'root': {
        'level': 'INFO',
        'handlers': ['console']
    }

})


class WordCountPipeline(Pipeline):
    def definition(self):
        self.add_ramp(WordRamp, 'sentence')
        self.add_intersection(SentenceSplitIntersection, 'sentence', 'word_to_kinesis')
        self.add_intersection(ExampleKinesisIntersection, 'word_to_kinesis')

        self.add_ramp(ExampleKinesisRamp, 'word')
        self.add_intersection(WordCountIntersection, 'word', 'word_count', grouper_cls=HashRingGrouper, processes=1)
        self.add_intersection(AggregateIntersection, 'word_count', grouper_cls=HashRingGrouper, processes=1)


WordCountPipeline().run()